<?php

namespace Drupal\lb_settings_report\Form;

use Drupal\Component\Diff\Diff;
use Drupal\Component\Utility\Crypt;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Config\StorageInterface;
use Drupal\Core\Diff\DiffFormatter;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\RendererInterface;
use Drupal\Core\Serialization\Yaml;
use Drupal\Core\Site\Settings;
use Drupal\Core\StringTranslation\TranslatableMarkup;
use Drupal\Core\Url;
use Drupal\layout_builder\Plugin\SectionStorage\OverridesSectionStorage;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a form for generating layout builder settings reports.
 */
class LayoutBuilderSettingsReportForm extends FormBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The config storage.
   *
   * @var \Drupal\Core\Config\StorageInterface
   */
  protected $configStorage;

  /**
   * The renderer.
   *
   * @var \Drupal\Core\Renderer\RendererInterface
   */
  protected $renderer;

  /**
   * The diff formatter.
   *
   * @var \Drupal\Core\Diff\DiffFormatter
   */
  protected $diffFormatter;

  /**
   * Constructs a new ConfigSingleImportForm.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Config\StorageInterface $config_storage
   *   The config storage.
   * @param \Drupal\Core\Renderer\RendererInterface $renderer
   *   The renderer service.
   * @param \Drupal\Core\Diff\DiffFormatter $diff_formatter
   *   The diff formatter service.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, StorageInterface $config_storage, RendererInterface $renderer, DiffFormatter $diff_formatter) {
    $this->entityTypeManager = $entity_type_manager;
    $this->configStorage = $config_storage;
    $this->renderer = $renderer;
    $this->diffFormatter = $diff_formatter;

    // Find all entity view displays.
    $options = $this->findConfiguration('entity_view_display');
    $target_types = [];

    // TODO: Add @var definition for these vars.
    // Filter out displays that don't have layout builder enabled.
    $this->subjectOptions = array_filter($options, function ($option) use (&$target_types) {
      $definition = $this->entityTypeManager->getDefinition('entity_view_display');
      $name = $definition->getConfigPrefix() . '.' . $option;
      $config = $this->configStorage->read($name);
      if (!empty($config['third_party_settings']['layout_builder'])) {
        $target_types[$config['targetEntityType']] = TRUE;
      }
      return !empty($config['third_party_settings']['layout_builder']);
    }, ARRAY_FILTER_USE_KEY);
    $type_keys = array_keys($target_types);
    $this->entityTargetTypes = [
      '' => $this->t('- Select entity type -'),
    ] + array_combine($type_keys, $type_keys);
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('config.storage'),
      $container->get('renderer'),
      $container->get('diff.formatter')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'lb_settings_report';
  }

  /**
   * Retrieves list of report types.
   *
   * @return array
   *   List of report types.
   */
  public function getReportTypes() {
    // TODO: Convert report types to plugins for scalability.
    return [
      'diff' => [
        'title' => $this->t('Diff'),
        'description' => $this->t('Provides a differential report for two different entities or entity types.'),
        'callback' => 'generateDiffReport',
      ],
      'single' => [
        'title' => $this->t('Single'),
        'description' => $this->t('Provides a layout builder settings report for a single entity or entity type.'),
        'callback' => 'generateSingleReport',
      ],
    ];
  }

  /**
   * Retrieves a list of options for the type dropdown.
   *
   * @return array
   *   List of options containing valid report types.
   */
  protected function getReportTypeOptions() {
    return array_map(function ($type) {
      return $type['title'];
    }, $this->getReportTypes());
  }

  /**
   * Retrieves the default subject field render array.
   *
   * @param string $name
   *   Name of current subject.
   *
   * @return array
   *   Render array.
   */
  protected function getDefaultSubjectField($name) {
    return [
      '#type' => 'select',
      '#title' => $this->t('Subject'),
      '#options' => ['' => ''],
      '#prefix' => "<div id='edit-{$name}-subject-wrapper'>",
      '#suffix' => '</div>',
      '#validated' => TRUE,
      '#states' => [
        'visible' => [
          ":input[name='{$name}[subject_type]']" => ['value' => 'entity_type'],
        ],
      ],
    ];
  }

  /**
   * Generates a subject field.
   */
  protected function generateSubjectField(&$form, $id, $name, $valid_types = NULL) {
    // If valid types isn't set, use full list.
    if (!isset($valid_types)) {
      $valid_types = array_keys($this->getReportTypes());
    }

    // Generate the state array.
    $state_array = [
      ':input[name="report_type"]' => array_map(function ($type) {
        return ['value' => $type];
      }, $valid_types),
    ];

    $form[$name] = [
      '#title' => $this->t('Subject @id', ['@id' => $id]),
      '#type' => 'fieldset',
      '#tree' => TRUE,
      '#states' => [
        'visible' => $state_array,
      ],
      'subject_type' => [
        '#title' => $this->t('Subject type'),
        '#type' => 'select',
        '#options' => [
          '' => $this->t('- Select a subject type -'),
          'entity' => $this->t('Overridden Entity'),
          'entity_type' => $this->t('Entity Type Display'),
        ],
        '#states' => [
          'required' => $state_array,
        ],
      ],
      'subject' => [
        '#type' => 'select',
        '#title' => $this->t('Subject'),
        '#options' => $this->subjectOptions,
        '#prefix' => "<div id='edit-{$name}-subject-wrapper'>",
        '#suffix' => '</div>',
        '#validated' => TRUE,
        '#states' => [
          'visible' => [
            ":input[name='{$name}[subject_type]']" => ['value' => 'entity_type'],
          ],
        ],
      ],
      'entity_type' => [
        '#type' => 'select',
        '#title' => $this->t('Subject entity type'),
        '#options' => $this->entityTargetTypes,
        '#validated' => TRUE,
        '#ajax' => [
          'callback' => '::updateEntityType',
          'wrapper' => "edit-{$name}-subject-entity-wrapper",
        ],
        '#states' => [
          'visible' => [
            ":input[name='{$name}[subject_type]']" => ['value' => 'entity'],
          ],
        ],
      ],
      'entity_id' => [
        '#type' => 'entity_autocomplete',
        '#title' => $this->t('Subject entity'),
        '#target_type' => 'user',
        '#selection_handler' => 'lb_settings_report:lb_override',
        '#prefix' => "<div id='edit-{$name}-subject-entity-wrapper'>",
        '#suffix' => '</div>',
        '#states' => [
          'visible' => [
            ":input[name='{$name}[entity_type]']" => ['empty' => FALSE],
          ],
        ],
      ],
    ];
  }

  /**
   * Provides HTML containing report type definitions.
   */
  protected function getReportTypeDescriptionMarkup() {
    $html = '<dl>';
    $types = $this->getReportTypes();
    array_walk($types, function ($type) use (&$html) {
      $html .= '<dt>' . $type['title'] . '</dt>';
      $html .= '<dd>' . $type['description'] . '</dt>';
    });
    $html .= '</dl>';
    return $html;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state, $config_type = NULL, $config_name = NULL) {
    $form['#attached']['library'][] = 'lb_settings_report/report';
    $form['report_type'] = [
      '#title' => $this->t('Report type:'),
      '#description' => $this->getReportTypeDescriptionMarkup(),
      '#type' => 'select',
      '#options' => $this->getReportTypeOptions(),
      '#default_value' => 'diff',
    ];
    // Add our subjects.
    $this->generateSubjectField($form, 1, 'subject1');
    $this->generateSubjectField($form, 2, 'subject2', ['diff']);

    $form['submit'] = [
      '#type' => 'button',
      '#value' => $this->t('Generate Report'),
      '#ajax' => [
        'callback' => '::generateReport',
      ],
    ];

    $form['report'] = [
      '#type' => 'markup',
      '#markup' => '<div class="lbSettingsReport__message"></div>',
    ];
    return $form;
  }

  /**
   * Ajax call after updating the entity type field.
   */
  public function updateEntityType($form, FormStateInterface $form_state) {
    $element = $form_state->getTriggeringElement();
    $element_name = $element['#name'];
    $parent_id = $element['#parents'][0];

    // Traverse into the correct field.
    $field = $form;
    while (count($element['#parents']) > 1 && $idx = array_shift($element['#parents'])) {
      $field = $field[$idx];
    }
    if (!empty($field['entity_id'])) {
      $field = $field['entity_id'];
      $target_type = $form_state->getValue($parent_id)['entity_type'];
      if (!empty($target_type)) {
        $field['#target_type'] = $target_type;

        // Store the selection settings in the key/value store and pass a hashed
        // key in the route parameters.
        // @see \Drupal\Core\Entity\Element\EntityAutocomplete::processEntityAutocomplete()
        $selection_settings = isset($field['#selection_settings']) ? $field['#selection_settings'] : [];
        $data = serialize($selection_settings) . $field['#target_type'] . $field['#selection_handler'];
        $selection_settings_key = Crypt::hmacBase64($data, Settings::getHashSalt());

        $key_value_storage = \Drupal::keyValue('entity_autocomplete');
        if (!$key_value_storage->has($selection_settings_key)) {
          $key_value_storage->set($selection_settings_key, $selection_settings);
        }

        $field['#autocomplete_route_parameters']['selection_settings_key'] = $selection_settings_key;
        $field['#autocomplete_route_parameters']['target_type'] = $target_type;
        $url = Url::fromRoute($field['#autocomplete_route_name'], $field['#autocomplete_route_parameters'])->toString(TRUE);
        $field['#attributes']['data-autocomplete-path'] = $url->getGeneratedUrl();
      }
    }
    else {
      throw new \Exception('Cannot find entity id field.');
    }
    return $field;
  }

  /**
   * Handles switching the configuration type selector.
   */
  protected function findConfiguration($config_type) {
    $names = [];

    // For a given entity type, load all entities.
    $entity_storage = $this->entityTypeManager->getStorage($config_type);
    foreach ($entity_storage->loadMultiple() as $entity) {
      $entity_id = $entity->id();
      if ($label = $entity->label()) {
        $names[$entity_id] = new TranslatableMarkup('@label (@id)', [
          '@label' => $label,
          '@id' => $entity_id,
        ]);
      }
      else {
        $names[$entity_id] = $entity_id;
      }
    }

    return $names;
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    // TODO: Build ajax-less fallback.
  }

  /**
   * Loads an entity type subject.
   *
   * @param string $subject
   *   Subject.
   *
   * @return array
   *   Subject array.
   */
  private function loadEntityTypeSubject($subject) {
    $definition = $this->entityTypeManager->getDefinition('entity_view_display');
    $name = $definition->getConfigPrefix() . '.' . $subject;
    $config = $this->configStorage->read($name);
    if (!empty($config['third_party_settings']['layout_builder'])) {
      return $config['third_party_settings']['layout_builder'];
    }
  }

  /**
   * Loads an entity subject.
   *
   * @param string $entity_type
   *   Entity type.
   * @param mixed $entity_id
   *   Entity id.
   *
   * @return array
   *   Subject array.
   */
  private function loadEntitySubject($entity_type, $entity_id) {
    $ret = [];
    $entity = $this->entityTypeManager->getStorage($entity_type)->load($entity_id);
    if ($entity) {
      if ($entity->hasField(OverridesSectionStorage::FIELD_NAME)) {
        $overrides = $entity->get(OverridesSectionStorage::FIELD_NAME);
        $ret['allow_custom'] = TRUE;
        $ret['enabled'] = TRUE;
        $ret['sections'] = [];
        foreach ($overrides->getSections() as $section) {
          $section_details = [
            'layout_id' => $section->getLayoutId(),
            'layout_settings' => $section->getLayoutSettings(),
          ];
          $components = $section->getComponents();
          if (!empty($components)) {
            $section_details['components'] = [];
            foreach ($components as $component) {
              $section_details['components'][$component->getUuid()] = $component->toArray();
            }
          }
          $ret['sections'][] = $section_details;
        }
      }
    }
    return $ret;
  }

  /**
   * Loads a subject based on the a specified subject array.
   *
   * @param array $subject_array
   *   The subject array.
   */
  private function loadSubjectByArray(array $subject_array) {
    switch ($subject_array['subject_type']) {
      case 'entity_type':
        return $this->loadEntityTypeSubject($subject_array['subject']);

      case 'entity':
        return $this->loadEntitySubject($subject_array['entity_type'], $subject_array['entity_id']);
    }
    return [];
  }

  /**
   * Attempts to generate the report.
   */
  public function generateReport($form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $report_types = $this->getReportTypes();
    $report_type = $values['report_type'];
    $html = $this->t('Could not generate report. Please check settings and try again.');
    if (!empty($report_types[$report_type]['callback'])) {
      $callback = $report_types[$report_type]['callback'];
      if (is_callable([$this, $callback])) {
        try {
          $html = call_user_func_array([$this, $callback], [$values]);
        }
        catch (Exception $e) {
        }
      }
    }
    $response = new AjaxResponse();
    $response->addCommand(new HtmlCommand('.lbSettingsReport__message', $html));
    return $response;
  }

  /**
   * Generates a single report.
   *
   * @param array $values
   *   Form values array.
   *
   * @return string
   *   HTML response.
   */
  public function generateSingleReport(array $values) {
    $subject = $this->loadSubjectByArray($values['subject1']);
    return '<pre class="lbSettingsReport__preformatted">' . Yaml::encode($subject) . '</pre>';
  }

  /**
   * Generates a diff report.
   *
   * @param array $values
   *   Form values array.
   *
   * @return string
   *   HTML response.
   */
  public function generateDiffReport(array $values) {
    // Retrieve and encode the subjects and then create diff.
    $subject1 = $this->loadSubjectByArray($values['subject1']);
    $subject2 = $this->loadSubjectByArray($values['subject2']);
    $left = explode(PHP_EOL, Yaml::encode($subject1));
    $right = explode(PHP_EOL, Yaml::encode($subject2));
    $diff = new Diff($left, $right);

    // Build diff table.
    $build = [];
    $build['#title'] = $this->t('Diff Report');
    $build['#attached']['library'][] = 'system/diff';
    $build['diff'] = [
      '#type' => 'table',
      '#attributes' => [
        'class' => ['diff'],
      ],
      '#header' => [
        ['data' => t('Subject 1'), 'colspan' => '2'],
        ['data' => t('Subject 2'), 'colspan' => '2'],
      ],
      '#rows' => $this->diffFormatter->format($diff),
    ];

    // Return rendered diff table.
    return $this->renderer->render($build);
  }

}
